; TFORTH
; A simple DOS Forth by Robert Oestling
; http://robos.org -- robost(at)home(dot)se
;
; 2004-11-30: Introducing hashed dictionaries and migrating to FASM.
; 2004-12-05: Done!
;
; A dictionary consists of 64 lists of headers. Hash code is calculated as
; name length*4 + first char*2 + second char*4. If there is no second char,
; it is assumed to be 0.
;
; Header structure.
;
; 00    16 b    Name (including count byte, zero padded).
; 10    word    Last header in chain.
; 12    word    Pointer to code.
; 14    byte    Flags.
;
; Flags.
;
; 0             set = immediate.
;
; Register usage.
;
; ss:sp         Data stack pointer (SP).
; ss:bp         Return stack pointer (RP).
; ds:si         Interpretation pointer (IP).
; bx            TOS.

use16

rstack_size = 0x40*2
dstack_size = 0x40*2
fstack_size = 0x10*2

include 'macros.asm'

        org     0x100
        jmp     start
core:
        times 64 dw 0
start:
        cld

        mov     ah,0x4a                 ; Resize our memory block.
        mov     bx,0x1000               ; 64kB
        int     0x21

        mov     word[fstack+fstack_size-2],0
        mov     byte[linebuf],0

        mov     bp,rstack+rstack_size
        mov     sp,dstack+dstack_size
        mov     si,code_bye

        mov     word[mode],mode_interpret
        mov     word[base],0x10
        mov     word[fptr],fstack+fstack_size-2
        mov     word[bptr],0
        jmp     code_quit

        xor     bx,bx

header  code_quit,"quit",flag_none
code_quit:
        call    code_do_ref
        dw      code_default_quit

code_default_quit:
        call    nest
        dw      code_do_lit,.mainf
        dw      code_fload_file         ; fload main.f
.next:  dw      code_read,code_word     ; Read a word.
        dw      code_process            ; Process it.
        dw      code_drop               ; We don't want to know the result.
        dw      code_do_branch,.next    ; Repeat.
.mainf  db      "main.f",0

; You can NOT do without these:
include 'core.asm'
include 'io.asm'
include 'string.asm'
include 'fload.asm'
include 'compiler.asm'

; You can do without these:
include 'mem.asm'
include 'gfx.asm'
include 'mouse.asm'

; 0x10 var base
header  code_base,"base",flag_none
code_base:
        call    code_do_constant
base:
        dw      0x10

; nstack.end var nptr
header  code_nptr,"nptr",flag_none
code_nptr:
        call    code_do_constant
nptr    dw      nstack.last

nstack:
        times   0x0f dw 0
.last:  dw      core
.end:

lastword dw     0
mode    dw      mode_interpret
here    dw      forthheap
fptr    dw      fstack+fstack_size-2
bptr    dw      0

rstack  rb    rstack_size
dstack  rb    dstack_size
fstack  rb    fstack_size

linebuf rb    0x80
tmpbuf  rb    0x20
wordbuf rb    0x20

forthheap:
